==> plib-0.1.tar.gz: plib-0.1/plib/gui/_widgets/_helpers.py <==
#!/usr/bin/env python
"""
Module HELPERS -- GUI Widget Helper Classes
Sub-Package GUI.WIDGETS of Package PLIB -- Python GUI Framework
Copyright (C) 2008 by Peter A. Donis

Defines "helper" classes that implement basic functionality used
by various widgets.
"""

from plib.stdlib import abstractlist, basesequence, normalize, SortMixin

class _PHelperData(object):
    """
    Abstract base class for helper items that update displayed
    text from data.
    """
    
    def __init__(self, data=None):
        if data is not None:
            self._update(data)
    
    def _update(self, data):
        """ Override to populate widget from data. """
        raise NotImplementedError

class _PHelperItem(_PHelperData):
    """
    Abstract mixin class for helper items. See the PHelper docstring
    below for more details.
    """
    
    def __init__(self, index, data=None):
        self._i = index
        _PHelperData.__init__(self, data)

class _PColsMixin(basesequence):
    """
    Mixin class to expose multiple columns of a single row
    or node in a multi-column widget as a sequence. Designed to
    be a mixin or helper for a _PHelperItem subclass. Assumes
    that data is a sequence of strings.
    """
    
    def _update(self, data):
        self._len = len(data)
        for index, item in enumerate(data):
            self[index] = item
    
    def __len__(self):
        return self._len

class _PHelperColLabels(_PColsMixin, _PHelperData):
    """
    Encapsulates the column labels of a multi-column widget
    with _PHelper.
    """
    
    def __init__(self, helper, labels=None):
        self._helper = helper
        _PHelperData.__init__(self, labels)
    
    def _get_data(self, index):
        return self._get_label(index)
    
    def _set_data(self, index, value):
        self._set_label(index, str(value))
        self._set_width(index, value.width)
        self._set_align(index, value.align)
        self._set_readonly(index, value.readonly)
    
    def _get_label(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_label(self, index, label):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_width(self, index, width):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_align(self, index, align):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_readonly(self, index, readonly):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError

class _PShift(object):
    """
    Mixin class to provide 'index-shift' functionality for
    implementation of sequence classes. Uses a dict with
    index values as keys as the underlying data store.
    """
    
    def __init__(self):
        self._items = {}
    
    def __len__(self):
        return len(self._items)
    
    def _shiftright(self, index):
        """ Move items at index and above 'up' one index (to be called
        before insertion at index). Note that this only re-indexes the
        items dict -- if there are any other side effects that should
        happen as a result of re-indexing, this method must be overridden
        to implement them. """
        
        n = len(self._items)
        if index < n:
            for i in range(n, index, -1):
                self._items[i] = self._items[i - 1]
            
            # List will be longer by one when insert/append occurs
            # in _add method below (so this statement isn't really
            # needed, it's there for symmetry with _shiftleft)
            del self._items[index]
    
    def _shiftleft(self, index):
        """ Move items above index in the list 'down' one index (to be
        called after a deletion at index). Note that this only re-indexes
        the items dict -- if there are any other side effects that should
        happen as a result of re-indexing, this method must be overridden
        to implement them. """
        
        n = len(self._items)
        if index < n:
            for i in range(index, n):
                self._items[i] = self._items[i + 1]
            
            # Shorten the list by one (item at n is stored at n - 1)
            # (note that unlike the del in _shiftright above, this
            # statement really is necessary to shorten the list)
            del self._items[n]

class _PHelper(_PShift, SortMixin, abstractlist):
    """
    Abstract class to provide common methods to make a widget look
    like a list of 'child' items. Each 'child' item should be a mixin
    of a _PHelperItem (see above) and either a child widget (such as a
    PListViewItem child of a PListView) or an object that manages a
    portion of the parent widget (such as a PTableRow for managing a
    single row of a PTable).
    
    Note that the basic method of implementation for this class is to
    override key list methods to provide "hooks" for side effects to
    occur when items are added or deleted. The "hook" methods must then
    be overridden in derived/mixin classes to implement the specific
    side effects desired (such as adding/deleting items from a list
    view or rows from a table).
    
    Note also that all of the sequence emulation methods work with the
    actual 'values', not the helper items. To access the helper items,
    use the items attribute (which is a standard dict mapping index
    values to the helper items).
    """
    
    def __init__(self, labels=None, data=None):
        _PShift.__init__(self)
        if labels is not None:
            self.labels = self.labelsclass(self, labels)
        abstractlist.__init__(self, data)
    
    def _newitem(self, index, value):
        """
        Derived classes can override to change the signature of
        the item class or substitute another callable that returns
        an item instance.
        """
        
        return self.itemclass(self, index, value)
    
    def _helperadd(self, index, value):
        """
        Derived classes should override to implement side effects of
        insertion (if they cannot be handled completely by the helper
        item's constructor).
        """
        
        pass
    
    def _helperdel(self, index, item):
        """
        Derived classes should override to implement side effects of
        deletion (if they cannot be handled completely by the helper item's
        destructor).
        """
        
        pass
    
    def _get(self, index):
        """
        Return the helper item at index. Note that we assume that each
        item stored in the items dict is a _PHelperItem or derived from it.
        """
        
        return self._items[index]
    
    def _set(self, index, value):
        """
        Update helper item at index with new value. Note that this method
        should never be called with an index out of range (meaning that the
        index will always be in the items dict); adding a new item should always
        be done using the _add method.
        """
        
        self._items[index]._update(value)
    
    def _add(self, index, value):
        """ Insert or append a helper item wrapped around value. """
        
        self._shiftright(index)
        self._helperadd(index, value)
        self._items[index] = self._newitem(index, value)
    
    def _del(self, index):
        """
        Delete the helper item at index. Note that we allow side effects
        to take place with deletion by calling _helperdel, but unless that
        method explicitly does so, the object wrapped by the deleted helper
        item is not deleted unless the helper item holds the only reference.
        """
        
        item = self._items[index]
        del self._items[index]
        self._helperdel(index, item)
        self._shiftleft(index)
    
    def __getitem__(self, index):
        if isinstance(index, int):
            return self._get(normalize(self.__len__(), index))
        else:
            raise TypeError, "List index must be an int."
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            self._set(normalize(self.__len__(), index), value)
        elif isinstance(index, slice) and (index.stop == index.start):
            # We're inserting an item or extending by a sequence
            # (e.g., from the constructor)
            for i, item in enumerate(value):
                self._add(index.start + i, item)
        else:
            raise TypeError, "List index must be an int."
    
    def __delitem__(self, index):
        if isinstance(index, int):
            self._del(normalize(self.__len__(), index))
        else:
            raise TypeError, "List index must be an int."

class _PRowHelperItem(_PColsMixin, _PHelperItem):
    """
    Helper item that exposes a sequence of columns, intended
    to map to a row in a multi-column widget which does *not*
    have nested tree functionality, such as a table.
    """
    
    def __init__(self, helper, index, data=None):
        self._helper = helper
        _PHelperItem.__init__(self, index, data)

class _PRowHelper(_PHelper):
    """
    Helper class for list or table-type widgets where each
    row has a helper item, but no nested tree structure.
    """
    
    def _helperadd(self, index, value):
        self._insert_row(index)
    
    def _helperdel(self, index, item):
        self._remove_row(index)
    
    def _insert_row(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _remove_row(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def rowcount(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_row(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_item(self):
        r = self.current_row()
        if r < 0:
            return None
        else:
            return self._items[r]

class _PTreeHelper(_PHelper):
    """
    Helper class for tree-type widgets, provides methods to
    easily manage the entire tree with this object as root.
    """
    
    def current_item(self):
        """ Derived classes must override to return the currently selected helper item. """
        raise NotImplementedError
    
    def set_current_item(self, item):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_index(self):
        return self.index(self.current_item())
    
    def set_current_index(self, index):
        self.set_current_item(self._items[index])

class _PTreeHelperItemCols(_PColsMixin, _PHelperData):
    """
    Class to expose columns of a tree view node as a
    sequence.
    """
    
    def __init__(self, helperitem, data=None):
        self._helperitem = helperitem
        _PHelperData.__init__(self, data)

class _PTreeHelperItem(_PHelperItem, _PTreeHelper):
    """
    Helper item that appears as its own nested sequence of
    helper items, intended to map to a tree-style widget such
    as a list view item. The data parameter must be a 2-tuple,
    with the first item being this item's strings (may be more
    than one because of columns), and the second item being a
    sequence of child data 2-tuples. The sequence of column
    strings is also exposed as the cols attribute.
    """
    
    colsclass = None
    
    def __init__(self, parent, index, data=None):
        self._parent = parent
        if data is not None:
            cols, children = data
        else:
            cols = children = None
        if not hasattr(self.__class__, 'itemclass'):
            self.__class__.itemclass = self.__class__
        if (self.colsclass is not None) and not hasattr(self, 'cols'):
            self.cols = self.colsclass(self, cols)
        _PHelperItem.__init__(self, index) # data must be None (it's in cols)
        _PTreeHelper.__init__(self, None, children) # labels must be None

==> plib-0.2.1.tar.gz: plib-0.2.1/plib/gui/_widgets/_helpers.py <==
#!/usr/bin/env python
"""
Module HELPERS -- GUI Widget Helper Classes
Sub-Package GUI.WIDGETS of Package PLIB -- Python GUI Framework
Copyright (C) 2008 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines "helper" classes that implement basic functionality used
by various widgets.
"""

from plib.stdlib import abstractlist, basesequence, normalize, SortMixin

class _PHelperData(object):
    """
    Abstract base class for helper items that update displayed
    text from data.
    """
    
    def __init__(self, data=None):
        if data is not None:
            self._update(data)
    
    def _update(self, data):
        """ Override to populate widget from data. """
        raise NotImplementedError

class _PHelperItem(_PHelperData):
    """
    Abstract mixin class for helper items. See the PHelper docstring
    below for more details.
    """
    
    def __init__(self, index, data=None):
        self._i = index
        _PHelperData.__init__(self, data)

class _PColsMixin(basesequence):
    """
    Mixin class to expose multiple columns of a single row
    or node in a multi-column widget as a sequence. Designed to
    be a mixin or helper for a _PHelperItem subclass. Assumes
    that data is a sequence of strings.
    """
    
    def _update(self, data):
        self._len = len(data)
        for index, item in enumerate(data):
            self[index] = item
    
    def __len__(self):
        return self._len

class _PHelperColLabels(_PColsMixin, _PHelperData):
    """
    Encapsulates the column labels of a multi-column widget
    with _PHelper.
    """
    
    def __init__(self, helper, labels=None):
        self._helper = helper
        _PHelperData.__init__(self, labels)
    
    def _get_data(self, index):
        return self._get_label(index)
    
    def _set_data(self, index, value):
        self._set_label(index, str(value))
        self._set_width(index, value.width)
        self._set_align(index, value.align)
        self._set_readonly(index, value.readonly)
    
    def _get_label(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_label(self, index, label):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_width(self, index, width):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_align(self, index, align):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _set_readonly(self, index, readonly):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError

class _PShift(object):
    """
    Mixin class to provide 'index-shift' functionality for
    implementation of sequence classes. Uses a dict with
    index values as keys as the underlying data store.
    """
    
    def __init__(self):
        self._items = {}
    
    def __len__(self):
        return len(self._items)
    
    def _shiftright(self, index):
        """ Move items at index and above 'up' one index (to be called
        before insertion at index). Note that this only re-indexes the
        items dict -- if there are any other side effects that should
        happen as a result of re-indexing, this method must be overridden
        to implement them. """
        
        n = len(self._items)
        if index < n:
            for i in range(n, index, -1):
                self._items[i] = self._items[i - 1]
            
            # List will be longer by one when insert/append occurs
            # in _add method below (so this statement isn't really
            # needed, it's there for symmetry with _shiftleft)
            del self._items[index]
    
    def _shiftleft(self, index):
        """ Move items above index in the list 'down' one index (to be
        called after a deletion at index). Note that this only re-indexes
        the items dict -- if there are any other side effects that should
        happen as a result of re-indexing, this method must be overridden
        to implement them. """
        
        n = len(self._items)
        if index < n:
            for i in range(index, n):
                self._items[i] = self._items[i + 1]
            
            # Shorten the list by one (item at n is stored at n - 1)
            # (note that unlike the del in _shiftright above, this
            # statement really is necessary to shorten the list)
            del self._items[n]

class _PHelper(_PShift, SortMixin, abstractlist):
    """
    Abstract class to provide common methods to make a widget look
    like a list of 'child' items. Each 'child' item should be a mixin
    of a _PHelperItem (see above) and either a child widget (such as a
    PListViewItem child of a PListView) or an object that manages a
    portion of the parent widget (such as a PTableRow for managing a
    single row of a PTable).
    
    Note that the basic method of implementation for this class is to
    override key list methods to provide "hooks" for side effects to
    occur when items are added or deleted. The "hook" methods must then
    be overridden in derived/mixin classes to implement the specific
    side effects desired (such as adding/deleting items from a list
    view or rows from a table).
    
    Note also that all of the sequence emulation methods work with the
    actual 'values', not the helper items. To access the helper items,
    use the items attribute (which is a standard dict mapping index
    values to the helper items).
    """
    
    labelsclass = None
    
    def __init__(self, labels=None, data=None):
        _PShift.__init__(self)
        if (labels is not None) and (self.labelsclass is not None):
            self.labels = self.labelsclass(self, labels)
        else:
            self.labels = []
        abstractlist.__init__(self, data)
    
    def _newitem(self, index, value):
        """
        Derived classes can override to change the signature of
        the item class or substitute another callable that returns
        an item instance.
        """
        
        return self.itemclass(self, index, value)
    
    def _helperadd(self, index, value):
        """
        Derived classes should override to implement side effects of
        insertion (if they cannot be handled completely by the helper
        item's constructor).
        """
        
        pass
    
    def _helperdel(self, index, item):
        """
        Derived classes should override to implement side effects of
        deletion (if they cannot be handled completely by the helper item's
        destructor).
        """
        
        pass
    
    def _get(self, index):
        """
        Return the helper item at index. Note that we assume that each
        item stored in the items dict is a _PHelperItem or derived from it.
        """
        
        return self._items[index]
    
    def _set(self, index, value):
        """
        Update helper item at index with new value. Note that this method
        should never be called with an index out of range (meaning that the
        index will always be in the items dict); adding a new item should always
        be done using the _add method.
        """
        
        self._items[index]._update(value)
    
    def _add(self, index, value):
        """ Insert or append a helper item wrapped around value. """
        
        self._shiftright(index)
        self._helperadd(index, value)
        self._items[index] = self._newitem(index, value)
    
    def _del(self, index):
        """
        Delete the helper item at index. Note that we allow side effects
        to take place with deletion by calling _helperdel, but unless that
        method explicitly does so, the object wrapped by the deleted helper
        item is not deleted unless the helper item holds the only reference.
        """
        
        item = self._items[index]
        del self._items[index]
        self._helperdel(index, item)
        self._shiftleft(index)
    
    def __getitem__(self, index):
        if isinstance(index, int):
            return self._get(normalize(self.__len__(), index))
        else:
            raise TypeError, "List index must be an int."
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            self._set(normalize(self.__len__(), index), value)
        elif isinstance(index, slice) and (index.stop == index.start):
            # We're inserting an item or extending by a sequence
            # (e.g., from the constructor)
            for i, item in enumerate(value):
                self._add(index.start + i, item)
        else:
            raise TypeError, "List index must be an int."
    
    def __delitem__(self, index):
        if isinstance(index, int):
            self._del(normalize(self.__len__(), index))
        else:
            raise TypeError, "List index must be an int."

class _PRowHelperItem(_PColsMixin, _PHelperItem):
    """
    Helper item that exposes a sequence of columns, intended
    to map to a row in a multi-column widget which does *not*
    have nested tree functionality, such as a table.
    """
    
    def __init__(self, helper, index, data=None):
        self._helper = helper
        _PHelperItem.__init__(self, index, data)

class _PRowHelper(_PHelper):
    """
    Helper class for list or table-type widgets where each
    row has a helper item, but no nested tree structure.
    """
    
    def _helperadd(self, index, value):
        self._insert_row(index)
    
    def _helperdel(self, index, item):
        self._remove_row(index)
    
    def _insert_row(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def _remove_row(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def rowcount(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_row(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_item(self):
        r = self.current_row()
        if r < 0:
            return None
        else:
            return self._items[r]

class _PTreeHelper(_PHelper):
    """
    Helper class for tree-type widgets, provides methods to
    easily manage the entire tree with this object as root.
    """
    
    def current_item(self):
        """ Derived classes must override to return the currently selected helper item. """
        raise NotImplementedError
    
    def set_current_item(self, item):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def current_index(self):
        return self.index(self.current_item())
    
    def set_current_index(self, index):
        self.set_current_item(self._items[index])

class _PTreeHelperItemCols(_PColsMixin, _PHelperData):
    """
    Class to expose columns of a tree view node as a
    sequence.
    """
    
    def __init__(self, helperitem, data=None):
        self._helperitem = helperitem
        _PHelperData.__init__(self, data)

class _PTreeHelperItem(_PHelperItem, _PTreeHelper):
    """
    Helper item that appears as its own nested sequence of
    helper items, intended to map to a tree-style widget such
    as a list view item. The data parameter must be a 2-tuple,
    with the first item being this item's strings (may be more
    than one because of columns), and the second item being a
    sequence of child data 2-tuples. The sequence of column
    strings is also exposed as the cols attribute.
    """
    
    colsclass = None
    
    def __init__(self, parent, index, data=None):
        self._parent = parent
        if data is not None:
            cols, children = data
        else:
            cols = children = None
        if not hasattr(self.__class__, 'itemclass'):
            self.__class__.itemclass = self.__class__
        if (self.colsclass is not None) and not hasattr(self, 'cols'):
            self.cols = self.colsclass(self, cols)
        _PHelperItem.__init__(self, index) # data must be None (it's in cols)
        _PTreeHelper.__init__(self, None, children) # labels must be None
